import abc
import re
from typing import Optional, List, Dict

from verbformen_cli.models import (
    PartOfSpeech,
    Level,
    Noun,
    Verb,
    Adjective,
    SearchResult,
    NotFound,
    Declension,
    Conjugation,
)

from bs4 import BeautifulSoup


class ParseError(Exception):
    ...


class AbstractParser(abc.ABC):
    @abc.abstractmethod
    def parse_page(self, html: str) -> SearchResult:
        ...


class VerbformenParser(AbstractParser):
    def parse_page(self, html: str) -> SearchResult:
        soup = BeautifulSoup(html, "html.parser")
        search = self._parse_search(soup)
        if self._not_found(soup):
            return NotFound(**{"search": search})

        part_of_speech = self._parse_part_of_speech(soup)
        description = self._description_paragraph(soup, part_of_speech)
        definitions = {
            "definitions": self._parse_definitions(soup),
            "search": self._parse_search(soup),
        }
        if part_of_speech == PartOfSpeech.NOUN:
            return Noun(
                **self._extract_noun_data(description) | definitions,
                declensions=self._parse_declensions(soup, ["Singular", "Plural"]),
            )

        elif part_of_speech == PartOfSpeech.VERB:
            return Verb(
                **self._extract_verb_data(description) | definitions,
                conjugations=self._parse_conjugations(
                    soup, ["Present", "Imperfect", "Present Subj.", "Imperf. Subj."]
                ),
            )

        elif part_of_speech == PartOfSpeech.ADJECTIVE:
            return Adjective(
                **self._extract_adjective_data(description) | definitions,
                declensions=self._parse_declensions(
                    soup, ["Masculine", "Neutral", "Feminine", "Plural"]
                ),
            )
        else:
            raise ValueError()

    def _not_found(self, soup: BeautifulSoup) -> bool:
        content = soup.find("meta", attrs={"name": "description"})["content"]
        return "List of all words starting with the text" in content

    def _parse_part_of_speech(self, soup: BeautifulSoup) -> PartOfSpeech:
        meta = soup.find("meta", attrs={"name": "description"})["content"]
        if "Declension of noun" in meta:
            return PartOfSpeech.NOUN
        if "Conjugation German verb" in meta:
            return PartOfSpeech.VERB
        if "Declension and comparison of" in meta:
            return PartOfSpeech.ADJECTIVE
        raise ParseError(f"failed to parse meta: {meta}")

    def _description_paragraph(
        self, soup: BeautifulSoup, part_of_speech: PartOfSpeech
    ) -> str:
        """text of paragraph above the definition box"""

        if part_of_speech == PartOfSpeech.NOUN:
            regex = "declension of the noun"
        elif part_of_speech == PartOfSpeech.ADJECTIVE:
            regex = "declension of the adjective"
        elif part_of_speech == PartOfSpeech.VERB:
            regex = "conjugation of the verb"
        else:
            raise ValueError(f"unexpected PartOfSpeech: {part_of_speech}")

        return clean_whitespace(
            soup.find(text=re.compile(f".*{regex}.*")).parent.parent.get_text()
        )

    def _extract_noun_data(self, description) -> Dict[str, str]:
        regex = re.match(
            r"The declension of the noun (?P<text>\S+) is in singular genitive "
            r"(?P<genitive>\S+) and in the plural nominative (?P<plural>\S+). "
            r"The noun (?P=text) is declined with the declension endings "
            r"(?P<genitive_ending>\S+)/(?P<plural_ending>\S+?). "
            r"(?:It can also be used with other endings\. )?"
            r"(?:In the plural is an umlaut\. )?"
            r"(?:It does not form plurals\. )?"
            r"(?:In the plural forms of (?P<possible_plural>\S+?) are possible\. )?"
            r"The voice of (?P=text) is (?P<gender>[a-z]+) and the article "
            r'"(?P<article>(der)|(die)|(das))"\. '
            r"(?:The noun can also be used with other genus and other articles\. )?"
            r"Here you can not only inflect (?P=text) but also all German nouns\. "
            r"(?:The noun is part of the thesaurus of Zertifikat Deutsch respectivly "
            r"Level (?P<level>\w{2})\.)?",
            description,
        )
        if not regex:
            raise ParseError(f"Parse Error:\n{description}")
        return regex.groupdict()

    def _parse_declensions(
        self, soup: BeautifulSoup, titles: List[str]
    ) -> List[Declension]:
        return [
            Declension(**self._parse_verbformen_table(soup, title)) for title in titles
        ]

    def _parse_conjugations(
        self, soup: BeautifulSoup, titles: List[str]
    ) -> List[Conjugation]:
        return [
            Conjugation(**self._parse_verbformen_table(soup, title)) for title in titles
        ]

    def _parse_verbformen_table(self, soup, title: str) -> Dict[str, str]:
        """
        Parse declension or conjugation table under the header

        :param soup: BeautifulSoup of page
        :param title: title of the declension/conjugation (singular, masculine, present)
        :return:
        """
        rows = soup.find("h2", text=title).find_next_sibling("table").find_all("tr")

        d = {}
        for row in rows:
            th = row.find("th")
            if not th:
                key = row.find("td").get_text()
                value = " ".join([td.get_text() for td in row.find_all("td")[1:]])
            else:
                key = th["title"]
                value = " ".join([td.get_text() for td in row.find_all("td")])
            d[clean_whitespace(key)] = clean_whitespace(value)
        d["title"] = title
        return d

    def _extract_verb_data(self, description) -> Dict[str, str]:
        regex = re.match(
            r"^The conjugation of the verb (?P<text>\S+) is (?P<behavior>\S+)\. "
            r"Basic forms are (?P<present>[\S ]+), (?P<imperfect>[\S ]+) and "
            r"(?P<perfect>[\S ]+?)\. "  # non-greedy for optional next line
            r"(?:The stem vowels are ([\S ]+)\. )?"
            r"The auxiliary verb of (?P=text) is (?P<auxiliary_verb>\S+?)\. "
            r"(?:(?P<secondary_auxiliary_verb>\S+) can be used as well\. )?"
            r"(?:First syllable (?P<separable_prefix>[\S-]+) of "
            r"(?P=text) is separable\. )?"
            r"(?:Can also be used not separable\. )?"
            r"(?:Prefix (?P<nonseparable_prefix>[\S-]+) of (?P=text) is not"
            r" separable\. )?"
            r"The flection is in (?P<flection>\S+) and the use as (?P<use>\S+)\. "
            r"For a better understanding, countless examples of the verb (?P=text) are"
            r" available\. "
            r"For practicing and consolidating, there are also free worksheets for"
            r" (?P=text)\. "
            r"You can not just (?P=text) conjugate, but all German verbs\. "
            r"(?:The verb is part of the thesaurus of Zertifikat Deutsch respectivly"
            r" Level"
            r" (?P<level>\w{2})\.)?",
            description,
        )
        if not regex:
            raise ParseError(f"Parse Error:\n{description}")
        return regex.groupdict()

    def _extract_adjective_data(self, description):
        # incomparable adjectives
        if regex := re.match(
            r"The declension of the adjective (?P<text>\S+) uses the incomparable form"
            r" (?P=text). "
            r"The adjective has no forms for the comparative and superlative. "
            r"The adjective (?P=text) can be used both attributively in front of a noun"
            r" as well as predicative in conjunction with a verb."
            r"One can not only inflect and compare (?P=text), but all German"
            r" adjectives\.",
            description,
        ):
            return regex.groupdict() | {"is_comparable": False}

        # comparable adjectives
        if regex := re.match(
            r"The declension of the adjective (?P<text>\S+) uses these forms of the"
            r" comparison (?P=text),(?P<comparative>\S+),(?P<superlative>.+)\. "
            r"The endings for the comparison in the comparative and superlative are"
            r" (?P<comparative_ending>\S+)/(?P<superlative_ending>\S+). "
            r"The adjective (?P=text) can be used both attributively in front of a noun"
            r" as well as predicative in conjunction with a verb."
            r"One can not only inflect and compare (?P=text), but all German"
            r" adjectives\.",
            description,
        ):
            return regex.groupdict() | {"is_comparable": True}
        raise ParseError(f"Parse Error:\n{description}")

    def _parse_level(self, soup: BeautifulSoup) -> Optional[Level]:
        level = soup.find(title=re.compile("^Vocabulary Certificate"))
        if not level:
            return None
        return Level[clean_whitespace(level.text)]

    def _parse_definition(self, soup: BeautifulSoup) -> str:
        container = soup.find(class_="rAbschnitt")
        definition = clean_whitespace(container.select(".vGrnd.rCntr")[0].get_text())
        return definition

    def _parse_search(self, soup: BeautifulSoup):
        return clean_whitespace(soup.find("input", type="search")["value"])

    def _parse_grammar(self, soup: BeautifulSoup) -> str:
        return clean_whitespace(
            soup.find(class_="rAbschnitt").select(".vStm.rCntr")[0].get_text()
        )

    def _parse_definitions(self, soup: BeautifulSoup):
        # image next to english definitions, if they exist
        eng_marker = soup.find(class_="rAbschnitt").find("img", alt="English")
        if not eng_marker:
            return []
        return [clean_whitespace(x) for x in eng_marker.parent.get_text().split(",")]

    def _parse_notes(self, soup) -> List[str]:
        return [
            clean_whitespace(tag.get_text())
            for tag in soup.find(class_="rAbschnitt").select(".rInf.vLeg.rClear")
        ]


def clean_whitespace(text: str):
    if not text:
        return text
    return re.sub(r"\s+", " ", text).strip()



from typing import List

from rich.align import Align
from rich.console import Console
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from verbformen_cli.models import (
    Noun,
    Verb,
    Adjective,
    SearchResult,
    NotFound,
    Declension,
)

console = Console()
max_definitions = 5


def display_summary(result: SearchResult, include_tables: bool):
    if isinstance(result, NotFound):
        console.print(
            Panel(
                Text("No results found", justify="center"),
                title=result.search,
                expand=True,
            )
        )
        return
    elif isinstance(result, Noun):
        level = result.level.value if result.level else None
        gender = result.gender.capitalize()
        endings = f"Endings: {result.genitive_ending}/{result.plural_ending}"
        top_left = " • ".join([x for x in [level, gender, endings] if x])
        word = f"{result.article} {result.text}"
        details = f"{result.genitive} • {result.plural}"

        table = declension_table(result.declensions)

    elif isinstance(result, Verb):
        level = result.level.value if result.level else None
        auxiliary = f"{result.auxiliary_verb}" + (
            f" (also, {result.secondary_auxiliary_verb})"
            if result.secondary_auxiliary_verb
            else ""
        )
        top_left = f"{level} • {result.behavior} • {auxiliary}"
        word = result.text
        details = f"{result.present} • {result.imperfect} • {result.perfect}"
        table = conjugation_table(result.conjugations)
    elif isinstance(result, Adjective):
        top_left = (
            f"Endings: {result.comparative_ending}/{result.superlative_ending}"
            if result.is_comparable
            else ""
        )
        word = result.text
        details = (
            f"{result.text} • {result.comparative} • {result.superlative}"
            if result.is_comparable
            else f"{result.text} • -- • --"
        )
        table = declension_table(result.declensions)
    else:
        raise ValueError()
    group = Panel(
        Group(
            Text(top_left, style="dim"),
            Text(word, justify="center", style="bold"),
            Text(details, justify="center"),
            Text(""),
            Text(", ".join(result.definitions[0:max_definitions]), justify="center"),
        ),
        expand=False,
        title=result.search,
    )
    console.print(Align.center(group))
    if include_tables:
        console.print(Align.center(table))


def declension_table(declensions: List[Declension]) -> Table:
    t = Table()
    t.title = "Declensions"
    t.show_header = True
    t.add_column("")
    for d in declensions:
        t.add_column(d.title)

    t.add_row("Nominative", *[d.Nominative for d in declensions])
    t.add_row("Accusative", *[d.Accusative for d in declensions])
    t.add_row("Dative", *[d.Dative for d in declensions])
    t.add_row("Genitive", *[d.Genitive for d in declensions])
    return t


def conjugation_table(conjugations):
    t = Table()
    t.title = "Conjugations"
    t.show_header = True
    t.add_column("")
    for c in conjugations:
        t.add_column(c.title)
    t.add_row("ich", *[c.ich for c in conjugations])
    t.add_row("du", *[c.du for c in conjugations])
    t.add_row("er", *[c.er for c in conjugations])
    t.add_row("wir", *[c.wir for c in conjugations])
    t.add_row("ihr", *[c.ihr for c in conjugations])
    t.add_row("sie", *[c.sie for c in conjugations])
    return t



import abc
import pathlib
import urllib.parse

import requests

from verbformen_cli.models import PartOfSpeech


class DownloaderError(Exception):
    ...


class AbstractDownloader(abc.ABC):
    @abc.abstractmethod
    def download(self, url: str) -> str:
        ...


class Downloader(AbstractDownloader):
    def download(self, url: str) -> str:
        response = requests.get(url)
        if response.status_code != 200:
            raise DownloaderError()
        return response.text


class CachedDownloader(AbstractDownloader):
    def __init__(self, cache_path: pathlib.Path, delegate: AbstractDownloader):
        self.delegate = delegate
        self.cache_dir = cache_path
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def download(self, url: str) -> str:
        path = self.cache_dir / urllib.parse.quote(url, safe="")
        if not path.is_file():
            page = self.delegate.download(url)
            path.write_text(page)
        return path.read_text()


def create_search_url(german_word: str, part_of_speech: PartOfSpeech = None) -> str:
    if part_of_speech == PartOfSpeech.NOUN:
        base = "https://www.verbformen.com/declension/nouns/?w="
    elif part_of_speech == PartOfSpeech.VERB:
        base = "https://www.verbformen.com/conjugation/?w="
    else:
        base = "https://www.verbformen.com/?w="
    return f"{base}{german_word}"



from verbformen_cli import parsers
from verbformen_cli.downloaders import (
    AbstractDownloader,
    create_search_url,
    CachedDownloader,
    Downloader,
)
from verbformen_cli.models import SearchResult, PartOfSpeech
from verbformen_cli.parsers import AbstractParser
from verbformen_cli.settings import settings


class VerbformenClient:
    def __init__(self, downloader: AbstractDownloader, parser: AbstractParser):
        self.downloader = downloader
        self.parser = parser

    def search(
        self, german_word: str, part_of_speech: PartOfSpeech = None
    ) -> SearchResult:
        url = create_search_url(german_word, part_of_speech)
        html = self.downloader.download(url)
        result = self.parser.parse_page(html)
        return result

    @classmethod
    def default_client(cls):
        downloader = CachedDownloader(settings.cache_dir, Downloader())
        parser = parsers.VerbformenParser()
        return VerbformenClient(downloader, parser)



import click

from verbformen_cli import clients
from verbformen_cli.display import display_summary
from verbformen_cli.models import PartOfSpeech


@click.command()
@click.argument("german_word")
@click.option(
    "--verb", "hint_pos", flag_value=PartOfSpeech.VERB.value, help="hint this is a verb"
)
@click.option(
    "--noun", "hint_pos", flag_value=PartOfSpeech.NOUN.value, help="hint this is a noun"
)
@click.option("--include_tables")
def lookup(german_word: str, hint_pos: str = None, include_tables: bool = False):
    """
    Lookup a word in the verbformen.net dictionary.

    If the part of speech is ambiguous, specify it using --noun or --verb

    """
    part_of_speech_hint = PartOfSpeech[hint_pos.upper()] if hint_pos else None
    client = clients.VerbformenClient.default_client()
    result = client.search(german_word, part_of_speech_hint)
    display_summary(result, include_tables)



from abc import ABC
from enum import Enum
from typing import Optional, List, Union

from pydantic import BaseModel, Field


class PartOfSpeech(Enum):
    NOUN = "noun"
    VERB = "verb"
    ADJECTIVE = "adjective"
    ADVERB = "adverb"


class Level(Enum):
    A1 = "A1"
    A2 = "A2"
    B1 = "B1"
    B2 = "B2"
    C1 = "C1"
    C2 = "C2"


class NounForm(BaseModel):
    singular: str
    plural: str


class PageResult(BaseModel):
    search: str = Field(example="Schloss")
    text: str = Field(example="das Schloss")
    grammar: str = Field(example="Schlosses · Schlösser⁰")
    part_of_speech: PartOfSpeech = Field(example=PartOfSpeech.NOUN)
    level: Optional[Level] = Field(example=Level.A2)
    definitions: List[str] = Field(example=["bolt", "castle", "lock"])
    notes: List[str] = Field(example="⁰ Depends on meaning")
    summary: str = Field(description="paragraph above summary tile")


class Definition(BaseModel, ABC):
    search: str = Field(description="Search term")
    definitions: List[str]
    part_of_speech: PartOfSpeech
    text: str = Field(description="German dictionary word")
    level: Optional[Level] = Field(example=Level.A1, default=None)


class Declension(BaseModel):
    Nominative: str
    Accusative: str
    Dative: str
    Genitive: str
    title: str


class Noun(Definition):
    part_of_speech = PartOfSpeech.NOUN
    genitive: str
    plural: str
    genitive_ending: str
    plural_ending: str
    gender: str
    article: str

    declensions: List[Declension] = []


class Conjugation(BaseModel):
    title: str
    ich: str
    du: str
    er: str
    wir: str
    ihr: str
    sie: str


class Verb(Definition):
    part_of_speech = PartOfSpeech.VERB
    behavior: str = Field(example="regular")
    present: str = Field(example="holt")
    imperfect: str = Field(example="holte")
    perfect: str = Field(example="hat geholt")
    auxiliary_verb: str = Field(example="haben")
    secondary_auxiliary_verb: Optional[str] = Field(example="sein", default=None)
    flection: str = Field(example="Active")
    use: str = Field(example="Main")
    separable_prefix: str = Field(example="-fern for fernsehen", default=None)

    conjugations: List[Conjugation] = []


class Adjective(Definition):
    part_of_speech = PartOfSpeech.ADJECTIVE
    is_comparable: bool
    comparative: Optional[str] = None
    superlative: Optional[str] = None
    comparative_ending: Optional[str] = None
    superlative_ending: Optional[str] = None

    declensions: List[Declension] = []


class NotFound(BaseModel):
    search: str = Field(description="Search term")


SearchResult = Union[Definition, NotFound]



import verbformen_cli.clients as clients
import verbformen_cli.models as models
import verbformen_cli.display as display

Client = clients.VerbformenClient
PageResult = models.PageResult
PartOfSpeech = models.PartOfSpeech
Definition = models.Definition
Noun = models.Noun
Verb = models.Verb
Adjective = models.Adjective
NotFound = models.NotFound
display_summary = display.display_summary

__version__ = "0.1.0"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['verbformen_cli']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.3,<5.0.0',
 'click>=7.0,<8.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'rich>=10.7.0,<11.0.0',
 'xextract>=0.1.8,<0.2.0']

entry_points = \
{'console_scripts': ['verbformen = verbformen_cli.cli:lookup']}

setup_kwargs = {
    'name': 'verbformen-cli',
    'version': '0.2.1',
    'description': 'Verbformen CLI',
    'long_description': '## Verbformen CLI\n\nUnofficial python client for the German-English dictionary at verbformen.com. Verbformen provides detailed conjugations and declensions of German words, along with a useful summary card that includes: \n\n- verb endings,\n- noun gender,\n- noun endings,\n- adjective endings, and\n- definitions\n\n### Installation\n\n```bash\npip install verbformen-cli\n```\n\n### Usage\nFrom the terminal:\n```\n$ verbformen Wörterbuch \n           ╭────────── Wörterbuch ───────────╮\n           │ A1 • Neutral • Endings: es/ü-er │\n           │         das Wörterbuch          │\n           │  Wörterbuch(e)s • Wörterbücher  │\n           │                                 │\n           │  dictionary, lexicon, wordbook  │\n           ╰─────────────────────────────────╯\n\n$ verbformen nachschlagen\n   ╭───────────────── nachschlagen ──────────────────╮\n   │ B1 • irregular • haben (also, sein)             │\n   │                  nachschlagen                   │\n   │ schlägt nach • schlug nach • hat nachgeschlagen │\n   │                                                 │\n   │                     look up                     │\n   ╰─────────────────────────────────────────────────╯\n\n```\n\nor, in code:\n```python\nfrom verbformen_cli import Client, PartOfSpeech\n\nclient = Client.default_client()\nclient.search("essen")\n# {\n#   "search": "essen",\n#   "definitions": [\n#     "eat",\n#     "consume"\n#   ],\n#   "part_of_speech": "verb",\n#   "text": "essen",\n#   "behavior": "irregular",\n#   "present": "isst",\n#   "imperfect": "aß",\n#   "perfect": "hat gegessen",\n#   "auxiliary_verb": "haben",\n#   "flection": "Active",\n#   "use": "Main",\n#   "level": "A1"\n# }\n```\n',
    'author': 'Nicholas Miller',
    'author_email': 'njmiller6@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/millern/verbformen-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
